import importlib
import sys
import traceback
import warnings

from .util import full_class_name

# TODO
# check out
#   https://github.com/ionelmc/python-tblib
#   https://stackoverflow.com/questions/6126007/python-getting-a-traceback-from-a-multiprocessing-process
# about pickling Exceptions with tracebacks.

# See also: boltons.tbutils
# See also: package `eliot`: https://github.com/itamarst/eliot/blob/master/eliot/_traceback.py
# See also: package `pebble`


class RemoteException(Exception):
    '''
    This class is used to wrap a "regular" Exception object to add support
    for pickling the traceback (to some extent), so that when this object
    is passed from the originating process into another process and handled
    there, the printout (such as after being raised) contains useful traceback
    info.

    Example:

            try:
                a = 0 / 1
                return a
            except:
                return RemoteException()

        This is equivalent to

            try:
                a = 0 / 1
                return a
            except exception as e:
                return RemoteException(e)

        # In another process, this object, say `x`, is retrieved from a queue.
        # Then,

            if isinstance(x, RemoteException):
                raise x

    The printout looks as if the exception has occurred this way:

        raise RemoteException(...) from e

    where `e` is the original exception object as happened in the original
    process.

    One could `raise` this object in the original process (w/o any pickling).
    The printout is meaningful. However, this is not an intended use case.

    There are ways to make `RemoteException` unpickle to an object of the
    original exception type directly, and behave quite like the original
    object. However, there are scanarios where it is useful to recognize
    that an exception object is one from another process via the
    `RemoteException` mechanism. As a result, it's decided to keep the
    `RemoteException` class in both originating and receiving processes,
    and make it behave differently from the original class in a few places.

    To get the original exception object, reach for the `exc_value` attribute.

    Efforts are made to support pickling; a pair of`to_dict/from_dict` methods
    further enable using other serializations.
    One peculiar case is that the producing environment used a custom
    exception class that is not available in the consuming environment;
    this is handled by the pickling support and `to_dict/from_dict`.
    '''

    def __init__(self, exc: BaseException = None, /):
        if exc is None:
            exc_type, exc_value, tb = sys.exc_info()
        else:
            if type(exc) is type:
                exc = exc()
            exc_type = type(exc)
            exc_value = exc
            tb = exc.__traceback__

        self._exc_name = exc_type.__name__
        self._exc_fullname = full_class_name(exc_type)
        self._exc_repr = repr(exc_value)
        self._exc_str = str(exc_value)
        self._exc_args = exc_value.args

        tb = traceback.TracebackException(exc_type, exc_value, tb)
        tb = ''.join(tb.format(chain=True)).strip('\n')
        tb = [tb]
        if issubclass(exc_type, RemoteException):
            tb = exc_value._exc_tb + tb
        self._exc_tb = tb

        # TODO: how to use the __cause__ attribute with a `self.exc_value`
        # that has a traceback properly attached?

        # There's a special attribute __cause__, which is not in `self.__dict__`.
        # This attribute makes the `raise` printout look like this:
        #
        #   ....
        #
        #   The above exception was the direct cause of the following exception:
        #
        #   ....
        # We are not using this attribute. The difficulty lies in attaching
        # a traceback object to `self.exc_value`.
        # Instead, we directly control the printout via customizing sys.excepthook.

    def __repr__(self):
        return f"{self.__class__.__name__}({self._exc_repr})"

    def __str__(self):
        return f"{self._exc_name}: {self._exc_str}"

    @property
    def args(self):
        return (self.exc_value, )

    def to_dict(self):
        return self.__dict__

    @classmethod
    def from_dict(cls, state):
        obj = cls.__new__(cls)
        obj.__setstate__(state)
        return obj

    def __reduce__(self):
        return (Exception.__new__, (self.__class__,), self.to_dict())
        # Customize the first 2 elements to be sure `__init__` is not called
        # during unpickling, b/c the exception object expected by `__init__`
        # may be unavailable in the unpickling environment.
        # The 3rd element is the input to `__setstate__`.

        # `Exception.__reduce__` is defined, hence `__getstate__` is never called.
        # Don't try to define `__getstate__` in order to customize its pickling.

    @property
    def exc_type(self):
        fullname = self._exc_fullname
        try:
            if '.' not in fullname:
                exc_type = eval(fullname)
            else:
                names = fullname.split('.')
                if len(names) == 2 and names[0] == '__main__':
                    __main__ = sys.modules['__main__']
                    exc_type = getattr(__main__, names[1])
                else:
                    mod = '.'.join(names[:-1])
                    try:
                        m = sys.modules[mod]
                    except KeyError:
                        m = importlib.import_module(mod)
                    exc_type = getattr(m, names[-1])

        except (NameError, ModuleNotFoundError, AttributeError) as e:
            warnings.warn(str(e))
            exc_type = Exception

        return exc_type

    @property
    def exc_value(self):
        exc_type = self.exc_type
        if exc_type is Exception and self._exc_name != 'Exception':
            val = exc_type(self._exc_fullname, *self._exc_args)
        else:
            val = exc_type(*self._exc_args)
        return val

    def format(self) -> str:
        delim = "\n\nThe above exception was the direct cause of the following exception:\n\n"
        return delim.join(self._exc_tb)

    def print(self):
        print(self.format(), file=sys.stderr)


def exit_err_msg(obj, exc_type=None, exc_value=None, exc_tb=None):
    if exc_type is None:
        return
    if isinstance(exc_value, RemoteException):
        msg = "Exiting {} with exception: {}\n{}".format(
            obj.__class__.__name__, exc_type, exc_value,
        )
        msg = "{}\n\n{}\n\n{}".format(
            msg, exc_value.format(),
            "The above exception was the direct cause of the following exception:")
        msg = f"{msg}\n\n{''.join(traceback.format_tb(exc_tb))}"
        return msg


_excepthook_ = sys.excepthook


def _my_excepthook(type_, value, tb):
    # Customize the printout upon `raise RemoteException(...)`.
    # This works in the default Python console but may not work in other
    # interactive shells such as `ptpython`, `ipython`.
    #
    #  https://stackoverflow.com/q/1261668/6178706
    #
    if type_ is RemoteException:
        print("{}\n\n{}\n".format(
            value.format(),
            "The above exception was the direct cause of the following exception:"),
            file=sys.stderr)
    _excepthook_(type_, value, tb)


sys.excepthook = _my_excepthook



import logging
from typing import Union

import uvicorn
from asgiref.typing import ASGIApplication  # such as `starlette.applications.Starlette`


logger = logging.getLogger(__name__)


# About server shutdown:
#  https://github.com/encode/uvicorn/issues/742
#  https://stackoverflow.com/questions/58133694/graceful-shutdown-of-uvicorn-starlette-app-with-websockets

# See tests for examples of server shutdown.

# About socket 'backlog':
# https://stackoverflow.com/a/12340078
# https://stackoverflow.com/questions/36594400/what-is-backlog-in-tcp-connections
# http://veithen.io/2014/01/01/how-tcp-backlog-works-in-linux.html

# Adapted from `uvicorn.main.run`.
def make_server(
        app: Union[str, ASGIApplication],
        *,
        host='0.0.0.0',
        port: int = 8000,
        log_level: str = None,
        debug: bool = None,
        access_log: bool = None,
        loop='auto',
        workers: int = 1,
        backlog: int = 64,
        **kwargs,
):
    '''
    `app`: a `Starlette` instance or the import string for such
        an instance, like 'mymodule:app'.

    `loop`: if you encounter errors, esp if you need to use
        the event loop before calling this function, use 'none'.

        If you don't need to use the eventloop at all before
        calling this function, then it's OK to pass in
        `loop='auto'`. In that case, `uvicorn` will use `uvloop`
        if that package is installed (w/o creating a new loop);
        otherwise it will create a new `asyncio` native event loop
        and set it as the default loop.

    `workers`: when used for `mpservice.mpserver.Server`, this should be 1.

    If user has their own ways to config logging, then pass in
    `log_config=None` in `kwargs`.

    `backlog`: this is passed to asyncio `loop.create_server` (in `asyncio.base_events`,
    where default is 100), and in-turn to `socket.listen`. Don't make this large
    unless you know what you're doing.
    '''
    if log_level is None:
        log_level = logging.getLevelName(logger.getEffectiveLevel()).lower()
    assert log_level in ('debug', 'info', 'warning', 'error')

    if debug is None:
        debug = log_level == 'debug'
    else:
        debug = bool(debug)

    if access_log is None:
        access_log = debug
    else:
        access_log = bool(access_log)

    config = uvicorn.Config(
        app,
        host=host,
        port=port,
        access_log=access_log,
        debug=debug,
        log_level=log_level,
        loop=loop,
        workers=workers,
        backlog=backlog,
        reload=debug and isinstance(app, str),
        **kwargs)

    server = uvicorn.Server(config=config)

    return server


def run_app(app, **kwargs):
    server = make_server(app, **kwargs)
    return server.run()



'''`ServerProcess` provides a server running in one process,
to be called from other processes for shared data or functionalities.

User should subclass this class to implement the functionalities they need.

Usage:

  (1) Define a subclass,

      class MyServerProcess(ServerProcess):
          def do_x(self, x):
              ...
              y = ...
              return y

    The subclass can define `__init__` and helper methods as needed,
    plus public methods as service APIs.

  (2) in main process,

      obj = MyServerProcess.start(...)

  (3) pass `obj` to other processes

  (4) in other processes, cal public methods of `obj`, e.g.

      y = obj.do_x(123)

`obj` is NOT an instance of the class `MyServerProcess`.
It's a "proxy" object, which is like a reference to a
`MyServerProcess` object in the "server process".
All public methods (i.e. named without a leading '_')
of `MyServerProcess` can be used on this proxy object from other processes.
Input and output should all be picklable objects, since this cross-process
communiation.

When the proxy of one server object is passed to multiple processes,
the calls to its public methods (even the same method) from different processes
are concurrent. This will show benefit if the method is I/O bound.
I guess we can think of the method call in the server process as if it is
happening in separate threads.

When all references to this proxy object have
been garbage-collected, the server process is shut down.
Usually user doesn't need to worry about it.
In order to proactively shut down the server process,
delete (all references to) the proxy object.

`ServerProcess.start()` can be used multiple times
to create multiple server objects, which reside in diff
processes and are independent of each other.

Example use cases:

  (1) Worker processes need to call an external service with
      certain categorical input, i.e. some return values are
      repeated. Use a server process to call the external
      service and do some caching of the results; worker
      processes call this server process.

  (2) Worker processes all need to load a large dataset
      into memory for some look up. Instead of have a copy
      of this large dataset in each worker process, load
      a single copy in a server process, which provides
      lookup service for the worker processes.
'''

from multiprocessing.managers import BaseManager


class _MyManager(BaseManager):
    pass


class ServerProcess:
    # In one application, there should not be two subclasses
    # of `ServerProcess` with the same name,
    # because subclass registration is by the class name.

    @classmethod
    def start(cls, *args, ctx=None, **kwargs):
        '''
        `args` and `kwargs` are passed on to the `__init__`
        method of this class (implemented by a subclass as needed).
        The method `__init__` is executed in the process that hosts
        the real server object.
        '''
        if cls.__name__ not in _MyManager._registry:
            _MyManager.register(
                cls.__name__,
                cls,
            )
        manager = _MyManager(ctx=ctx)
        manager.start()  # pylint: disable=consider-using-with
        obj = getattr(manager, cls.__name__)(*args, **kwargs)
        return obj



'''
This module provides tools to use a "named pipe" to communicate between
two Python processes on the same machine.

Usually the two Python processes are two separately started programs.
If they are two processes created by `multiprocessing` in a single program,
then one would directly use `multiprocessing.Pipe` instead of this module.

One creates a `Server` object in one process and a `Client` object in
the other process, providing the same `path` argument.
The two objects can be created in any order, and their roles are symmetric.
The different names simply remind the application to create one of each
in the two processes.

Two uni-directional pipes are created between the two processes.
In each process, one "sends" (or writes) to one pipe and "receives" (or reads)
from the other pipe. The roles of the two pipes in the two processes are flipped.

The `send` and `recv` functions take and return picklable Python objects.
While `send` does not block as long as system buffer has space,
`recv` blocks until one data item is read.

It's up to the application to design conventions understood by both sides
to use special values to signal start, finish, and the like.

See `multiprocessing.connection.Connection` for documentation on the
`send` and `recv` methods.
'''

import os
import stat
from multiprocessing.connection import Connection


def _mkfifo(path: str):
    if os.path.exists(path):
        assert stat.S_ISFIFO(os.stat(path).st_mode), f"file '{path}' exists but is not a FIFO"
    else:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        os.mkfifo(path)


class _Pipe:
    def __init__(self, rpath: str, wpath: str):
        self._rpath = os.path.abspath(rpath)
        self._wpath = os.path.abspath(wpath)
        _mkfifo(self._rpath)
        _mkfifo(self._wpath)

        hw = os.open(self._wpath, os.O_SYNC | os.O_CREAT | os.O_RDWR)
        self._writer = Connection(hw, readable=False)
        self._reader = None

    def send_bytes(self, buf, offset=0, size=None):
        self._writer.send_bytes(buf, offset=offset, size=size)

    def send(self, obj):
        self._writer.send(obj)

    def _get_reader(self):
        if self._reader is None:
            # Open for reading will block until the other end
            # has opened the same path for writing.
            # That's why we don't open this in `__init__`.
            # In contrast, open for writing does not block.
            hr = os.open(self._rpath, os.O_RDONLY)
            self._reader = Connection(hr, writable=False)
        return self._reader

    def recv_bytes(self, maxlength=None):
        return self._get_reader().recv_bytes(maxlength)

    def recv_bytes_info(self, buf, offset=0):
        return self._get_reader().recv_bytes_into(buf, offset)

    def recv(self):
        return self._get_reader().recv()


class Server(_Pipe):
    def __init__(self, path: str):
        super().__init__(path + '.1', path + '.2')


class Client(_Pipe):
    def __init__(self, path: str):
        super().__init__(path + '.2', path + '.1')
